
package com.reactlibrary;

import android.annotation.SuppressLint;
import android.content.Context;
import android.os.AsyncTask;
import android.support.annotation.NonNull;
import android.util.Log;

import com.facebook.react.bridge.Callback;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.modules.core.DeviceEventManagerModule;
import com.iceteck.silicompressorr.CompressionException;
import com.iceteck.silicompressorr.SiliCompressor;
import com.iceteck.silicompressorr.VideoConversionProgressListener;

import java.io.File;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import javax.annotation.Nullable;

/**
 * Created by Amit Kumar Sahoo on 13th Oct 2020.
 * Email : amit.sahoo@mindfiresolutions.com
 * Module class that contains RN exposed methods
 */
public class RNVideoCompressionModule extends ReactContextBaseJavaModule {

    //Pre-defined compression level
    private static final String COMPRESSION_LOW = "LOW";
    private static final String COMPRESSION_MEDIUM = "MEDIUM";
    private static final String COMPRESSION_HIGH = "HIGH";

    //Emitter to listen the conversion progress
    private static final String VIDEO_COMPRESS_PROGRESS_EMITTER = "VIDEO_COMPRESS_PROGRESS_EMITTER";

    private final ReactApplicationContext reactContext;
    private Callback errorCallback;
    private Callback successCallback;

    //Tag
    private final String TAG = "SiliVideoCompressor";

    public RNVideoCompressionModule(ReactApplicationContext reactContext) {
        super(reactContext);
        this.reactContext = reactContext;
    }

    @Override
    public String getName() {
        return "CompressModule";
    }

    @Nullable
    @Override
    public Map<String, Object> getConstants() {
        final Map<String, Object> constants = new HashMap<>();
        constants.put(COMPRESSION_LOW, COMPRESSION_LOW);
        constants.put(COMPRESSION_MEDIUM, COMPRESSION_MEDIUM);
        constants.put(COMPRESSION_HIGH, COMPRESSION_HIGH);
        constants.put(VIDEO_COMPRESS_PROGRESS_EMITTER, VIDEO_COMPRESS_PROGRESS_EMITTER);
        return constants;
    }

    /**
     * Method to compress video in 3 levels
     *
     * @param inputFilePath      : input file path in String ( "/storage/emulated/0/Download/20201003_095653.mp4" )
     * @param compressionQuality : Constant HIGH, MEDIUM, LOW
     * @param errorCallback      : Error CallBack(String : errorMessage)
     * @param successCallback    : Success CallBack(String : fileName, String : size, String : Absolute path, String : message)
     */
    @ReactMethod
    public void compressVideo(@NonNull String inputFilePath, @NonNull String compressionQuality, @NonNull Callback errorCallback, @NonNull Callback successCallback) {
        this.errorCallback = errorCallback;
        this.successCallback = successCallback;
        String output = Utility.getOutputFilePath(inputFilePath, reactContext);

        switch (compressionQuality) {
            case COMPRESSION_LOW:
                new VideoCompressAsyncTask(this.reactContext,
                        0.1F, 0.5F).execute(inputFilePath, output);
                break;
            case COMPRESSION_MEDIUM:
                new VideoCompressAsyncTask(this.reactContext,
                        0.5F, 0.5F).execute(inputFilePath, output);
                break;
            case COMPRESSION_HIGH:
                new VideoCompressAsyncTask(this.reactContext,
                        1F, 0.5F).execute(inputFilePath, output);
                break;
            default:
                errorCallback.invoke("Invalid compress option selected");
        }
    }

    /**
     * Method to compress percentage wise to handle it, more precisely
     *
     * @param inputFilePath               : input file path in String ( "/storage/emulated/0/Download/20201003_095653.mp4" )
     * @param reduceVideoQualityToPercent : varies from 0-1 (like : 0.2F)
     * @param reduceHeightWidthToPercent  : varies from 0-1 (like : 0.2F)
     * @param errorCallback               : Error CallBack(String : errorMessage)
     * @param successCallback             : Success CallBack(String : fileName, String : size, String : Absolute path, String : message)
     */
    @ReactMethod
    public void compressVideoWithQuality(@NonNull String inputFilePath, @NonNull Float reduceVideoQualityToPercent,
                                         @NonNull Float reduceHeightWidthToPercent, @NonNull Callback errorCallback,
                                         @NonNull Callback successCallback) {

        this.errorCallback = errorCallback;
        this.successCallback = successCallback;
        String output = Utility.getOutputFilePath(inputFilePath, reactContext);

        new VideoCompressAsyncTask(this.reactContext,
                reduceVideoQualityToPercent, reduceHeightWidthToPercent).execute(inputFilePath, output);
    }

    /**
     * AsyncTask to handle background compression
     */
    @SuppressLint("StaticFieldLeak")
    class VideoCompressAsyncTask extends AsyncTask<String, Float, String> {

        private Context mContext;
        private Float amountToCompressToLocal;
        private Float reduceHeightWidthToPercent;

        public VideoCompressAsyncTask(@NonNull Context context, Float percentToCompressDownToLocal, Float reduceHeightWidthToPercent) {
            mContext = context;

            //set video quality in percentage
            if (percentToCompressDownToLocal == null) {
                this.amountToCompressToLocal = 0.5F;
            } else if (percentToCompressDownToLocal < 0.0 || percentToCompressDownToLocal > 1.0) {
                this.amountToCompressToLocal = 0.5F;
            } else {
                this.amountToCompressToLocal = percentToCompressDownToLocal;
            }

            //Set video size in percentage
            if (reduceHeightWidthToPercent == null) {
                this.reduceHeightWidthToPercent = 0.5F;
            } else if (reduceHeightWidthToPercent < 0.0 || reduceHeightWidthToPercent > 1.0) {
                this.reduceHeightWidthToPercent = 0.5F;
            } else {
                this.reduceHeightWidthToPercent = reduceHeightWidthToPercent;
            }
        }

        @Override
        protected void onPreExecute() {
            super.onPreExecute();
        }

        @Override
        protected void onProgressUpdate(Float... values) {
            if (values == null) {
                return;
            }
            if (values.length <= 0) {
                return;
            }
            try {
                if (values[0] != null) {
                    float flt = ((values[0]) * 100);
                    Log.d(TAG, "Progress Complete: " + (flt) + "%");
                    reactContext.getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
                            .emit(VIDEO_COMPRESS_PROGRESS_EMITTER, ((int) flt));

                    if (flt > 40) {
                        if (false) { //Flip this to true to foce manual cancel when progress hits 40%
                            Log.d("1", "Attempting manual cancel of progress conversion @ >= 40%");
                            SiliCompressor.with(this.mContext).cancelVideoCompression();
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
                errorCallback.invoke(e.getMessage());
            }
            super.onProgressUpdate(values);
        }

        @Override
        protected String doInBackground(String... paths) {
            String filePath = null;
            Log.d("d", "Do in background, amount to compress local == " + amountToCompressToLocal);

            try {
                //Old Method
//                filePath = SiliCompressor.with(mContext).compressVideo(mContext, paths[0], paths[1]);

                //New Method
                try {
                    filePath = SiliCompressor.with(mContext, true).compressVideo(new VideoConversionProgressListener() {
                        @Override
                        public void videoConversionProgressed(float progressPercentage, Long estimatedNumberOfMillisecondsLeft) {
                            publishProgress(progressPercentage);
                        }
                    }, paths[0], paths[1], this.amountToCompressToLocal, this.reduceHeightWidthToPercent);
                } catch (CompressionException ce) {
                    ce.printStackTrace();
                    errorCallback.invoke(ce.getMessage());
                }

            } catch (ArrayIndexOutOfBoundsException e) {
                e.printStackTrace();
                errorCallback.invoke(e.getMessage());
            }
            return (filePath == null) ? "" : filePath;
        }


        @Override
        protected void onPostExecute(String compressedFilePath) {
            super.onPostExecute(compressedFilePath);
            processCompressedData(compressedFilePath);
        }

        /**
         * method to process output file from the path and fetch usable file info
         *
         * @param filePath : output file absolute path
         */
        private void processCompressedData(String filePath) {
            try {
                File videoFile = new File(filePath);

                //File name
                String fileName = videoFile.getName();
                //File length
                float length = videoFile.length();
                //Success message
                String text = String.format(Locale.US, "%s\nName: %s\nSize: %s", "Complete", fileName, length);
                successCallback.invoke(fileName, length, filePath, text);
                Log.d(TAG, text);
            } catch (Exception e) {
                errorCallback.invoke(e.getMessage());
                e.printStackTrace();
            }
        }
    }
}
